from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from functools import singledispatchmethod
from django.db.models import Model
from django.template.loader import render_to_string
from .utils import _channel_name_for_instance

# Turbo Streams CRUD operations
APPEND = "append"
PREPEND = "prepend"
REPLACE = "replace"
UPDATE = "update"
REMOVE = "remove"
BEFORE = "before"
AFTER = "after"


class ModelBroadcast:
    def __init__(self, model):
        self.model = model
        super().__init__()
    pass


SELECTOR_CSS = 'css'
SELECTOR_ID = 'id'
SELECTOR_TYPES = (SELECTOR_ID, SELECTOR_CSS)


class Turbo:
    """
    A reference to a specific broadcast.
    Broadcast names can be provided as strings or can be generated by passing in a django instance.

    Eg:
        Turbo('broadcast_name')
        Turbo(django_instance)

    """

    @singledispatchmethod
    def __init__(self):
        # Override with singledispatchmethods
        raise NotImplemented

    @__init__.register
    def _init_with_str(self, stream_name: str = ''):

        self.stream_name = stream_name
        self.stream_target = None

    @__init__.register
    def _init_with_model(self, stream_target: Model):
        self.stream_target = stream_target
        self.stream_name = ''

    def update_stream_name(self):
        # Update the channel name with the stream target's latest pk if needed.
        if self.stream_target:
            self.stream_name = _channel_name_for_instance(self.stream_target)

    def render(self, template_name, context=None) -> "TurboRender":
        """
        Returns a TurboRender object from a django template.  This rendered template
        can then be broadcast to subscribers with the TurboRender actions
        (eg: append, update, etc...)

        Takes a template name and context identical to Django's render() method.
        """
        if context is None:
            context = {}

        self.update_stream_name()

        rendered_frame = render_to_string(template_name, context)
        return TurboRender(self.stream_name, rendered_frame)

    def render_from_string(self, rendered_template: str) -> "TurboRender":
        """
        Returns a TurboRender object from a string.
        """
        self.update_stream_name()

        return TurboRender(self.stream_name, rendered_template)

    def remove(self, selector=None, id=None):
        """
        Send a broadcast to remove an element from a turbo frame.
        """
        # Remove does not require a template so allow it to pass through without a render().
        self.update_stream_name()
        turboframe = TurboRender(self.stream_name, '')
        return turboframe.remove(selector=selector, id=id)


class TurboRender:
    """
    A rendered template, ready to broadcast using turbo.
    """

    def __init__(
        self,
        stream_name: str,
        rendered_template: str = '',
    ):
        self.stream_name = stream_name
        self.rendered_template = rendered_template

    def append(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, APPEND)

    def prepend(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, PREPEND)

    def replace(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, REPLACE)

    def update(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, UPDATE)

    def remove(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, REMOVE)

    def before(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, BEFORE)

    def after(self, selector=None, id=None):
        """Add a target action to the given selector."""
        return self._add_target(selector, id, AFTER)

    def _add_target(self, selector, id, action):
        if selector is None != id is None:
            raise ValueError("Either selector or id can be used as a parameter.")

        selector_type = SELECTOR_CSS
        if selector is None:
            selector_type = SELECTOR_ID
            selector = id

        return self._broadcast(
            self._render_frame(selector_type=selector_type, selector=selector, action=action)
        )

    def _render_frame(self, selector_type, selector, action) -> str:

        template_context = {
            "action": action,
            "use_css_selector": selector_type == SELECTOR_CSS,
            "selector": selector,
        }
        # Remove actions don't have contents, so only add context for model
        # template if it's not a remove action.
        if action != REMOVE:
            template_context["rendered_template"] = self.rendered_template

        rendered_frame = render_to_string("turbo/stream.html", template_context)
        return rendered_frame

    def _broadcast(self, rendered_frame):
        channel_layer = get_channel_layer()
        async_to_sync(channel_layer.group_send)(
            self.stream_name,
            {
                "type": 'notify',
                "channel_name": self.stream_name,
                "rendered_template": rendered_frame
            },
        )



from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer
from django.core.signing import Signer, BadSignature

signer = Signer()


class TurboStreamException(Exception):
    pass


class TurboStreamsConsumer(JsonWebsocketConsumer):
    def connect(self):
        self.accept()

    def notify(self, event):
        signed_channel_name = signer.sign(event["channel_name"])
        self.send_json(
            {
                "signed_channel_name": signed_channel_name,
                "data": event.get("rendered_template")
            }
        )

    def receive_json(self, content, **kwargs):
        try:
            channel_name = signer.unsign(content["signed_channel_name"])
        except (BadSignature, KeyError):
            raise TurboStreamException(
                "Signature is invalid or not present. This could be due to a misbehaving client."
            )
        message_type = content["type"]
        if message_type == "subscribe":
            async_to_sync(self.channel_layer.group_add)(channel_name, self.channel_name)
        elif message_type == "unsubscribe":
            async_to_sync(self.channel_layer.group_discard)(
                channel_name, self.channel_name
            )



from django.db.models import Model


def make_channel_name(model_label, pk):
    return f"BROADCAST-{model_label}-{pk}".lower()


def get_channel_name(identifier):
    if isinstance(identifier, Model):
        return _channel_name_for_instance(identifier)
    else:
        return identifier.__str__()


def _channel_name_for_instance(instance: Model):
    return make_channel_name(instance._meta.label, instance.pk)



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from .utils import get_channel_name

from django.utils.module_loading import autodiscover_modules

# Bring classes up to turbo namespace.
from .classes import (
    ModelBroadcast,
    Turbo,
    APPEND, PREPEND, REPLACE, UPDATE, REMOVE, BEFORE, AFTER,
)

default_app_config = "turbo.apps.TurboDjangoConfig"


def autodiscover():
    """
    Look for ModelBroadcast classes in broadcasts.py
    """
    autodiscover_modules('broadcasts')


def broadcast_stream(
    stream_target,
    action,
    template,
    context,
    dom_target=None,
    css_selector=None,
    send_type="notify",
    extra_payload=None,
):
    """
    Send a Broadcast to all Websocket Clients registered to a specific stream!
    Here for manual control of streams and backwards-compatibility.
    """
    if extra_payload is None:
        extra_payload = dict()

    if dom_target is None:
        raise ValueError(
            'Either dom_target or css_selector must be set as a parameter to broadcast_stream().')

    selector_type = "id"
    selector = dom_target

    if dom_target is None:
        selector_type = "css"
        selector = css_selector

    channel_layer = get_channel_layer()
    channel_name = get_channel_name(stream_target)
    async_to_sync(channel_layer.group_send)(
        channel_name,
        {
            "type": send_type,
            "action": action,
            "channel_name": channel_name,
            "template": template,
            "context": context,
            "selector_type": selector_type,
            "selector": selector,
            **extra_payload,
        },
    )



import turbo

class TurboMixin:

    """
    Mixin that attaches a .turbo attribute to a django model.
    This allows for any instance to broadcast to pages listening to the
    object channel.

    Instances must have a primary key assigned to be broadcast, or else a ValueError
    will be thown.

    Ex:
        room = Room.objects.first()
        room.turbo.render("chat/room_name.html", {"room": room}).replace(id="update-room")
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.turbo = turbo.classes.Turbo(self)



from django.db.models.signals import post_save, post_delete


def register(*models, site=None):
    """
    Register the given model(s) classes with the ModelBroadcast.
    This attaches the on_save and on_delete methods to the provided
    model signals.

    .. code-block:: python

        @turbo.register(Room)
        class RoomBroadcast(turbo.ModelBroadcast):

            def on_save(self, room, created, *args, **kwargs):
                    pass

            def on_delete(self, room, *args, **kwargs):
                    pass

    """
    from turbo.classes import ModelBroadcast

    def _model_broadcast_wrapper(broadcast_class):
        if not models:
            raise ValueError('At least one model must be passed to register.')

        if not issubclass(broadcast_class, ModelBroadcast):
            raise ValueError('Wrapped class must subclass ModelBroadcast.')

        for model in models:
            model.broadcast_class = broadcast_class(model)
            post_save.connect(post_save_broadcast_model, sender=model)
            post_delete.connect(post_delete_broadcast_model, sender=model)

        return broadcast_class
    return _model_broadcast_wrapper


def post_save_broadcast_model(sender, instance, **kwargs):
    if hasattr(sender.broadcast_class, 'on_save'):
        sender.broadcast_class.on_save(instance, kwargs)


def post_delete_broadcast_model(sender, instance, **kwargs):
    if hasattr(sender.broadcast_class, 'on_delete'):
        sender.broadcast_class.on_delete(instance, kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['turbo', 'turbo.templatetags']

package_data = \
{'': ['*'], 'turbo': ['static/turbo/js/*', 'templates/turbo/*']}

install_requires = \
['Django>=3.0.0', 'channels>=2.0.0']

setup_kwargs = {
    'name': 'turbo-django',
    'version': '0.2.1',
    'description': 'Integrate Hotwire Turbo with Django allowing for a Python-driven dynamic web experience.',
    'long_description': '[![Build Status](https://img.shields.io/endpoint.svg?url=https%3A%2F%2Factions-badge.atrox.dev%2Fhotwire-django%2Fturbo-django%2Fbadge%3Fref%3Dmain&style=flat)](https://actions-badge.atrox.dev/hotwire-django/turbo-django/goto?ref=main)\n[![Documentation Status](https://readthedocs.org/projects/turbo-django/badge/?version=latest)](https://turbo-django.readthedocs.io/en/latest/?badge=latest)\n[![Issues](https://img.shields.io/github/issues/hotwire-django/turbo-django)](https://img.shields.io/github/issues/hotwire-django/turbo-django)\n[![Twitter](https://img.shields.io/twitter/url?style=social&url=https%3A%2F%2Ftwitter.com%2FDjangoHotwire)](https://twitter.com/intent/tweet?text=Wow:&url=https%3A%2F%2Fgithub.com%2Fhotwire-django%2Fturbo-django)\n\n# Turbo for Django\n\n\nIntegrate [Hotwire Turbo](https://turbo.hotwired.dev/) with Django with ease.\n\n\n## Requirements\n\n- Python 3.6+\n- Django 3.1+\n- Channels 3.0+ _(Optional for Turbo Frames, but needed for Turbo Stream support)_\n\n##Installation\n\nTurbo Django is available on PyPI - to install it, just run:\n\n    pip install turbo-django\n\n_Note: Both Hotwire and this library are still in beta development and may introduce breaking API changes between releases.  It is advised to pin the library to a specific version during install._\n\n## Quickstart\nWant to see Hotwire in action? Here\'s a simple broadcast that can be setup in less than a minute.\n\n**The basics:**\n\n* A web page subscribes to a specific broadcast name.\n\n* A view sends a rendered template to all subscribed pages telling the page where to position the new content.\n\n\n### Example\n\nFirst, create a view that takes a broadcast name.\n\n\n```python\nfrom django.urls import path\nfrom django.views.generic import TemplateView\n\nurlpatterns = [\n    path(\'quickstart/\', TemplateView.as_view(template_name=\'broadcast_example.html\'))\n]\n```\n\n```html\n# broadcast_example.html\n\n{% load turbo_streams %}\n<!DOCTYPE html>\n<html lang="en">\n<head>\n    {% include "turbo/head.html" %}\n</head>\n<body>\n    {% turbo_subscribe \'broadcast_name\' %}\n\n    <p class="broadcast_box_class" id="broadcast_box">Placeholder for broadcast</p>\n</body>\n</html>\n```\n\nNow run ``./manage.py shell``.  Create a Turbo object that references the broadcast name.  Tell the object to render a ``TurboRender`` object from the string, and then broadcast a command to `update` the inside of the element with id `broadcast_box` on all subscribed pages.\n\n```python\nfrom turbo import Turbo\nfrom datetime import datetime\n\nTurbo(\'broadcast_name\').render_from_string(\n    f"{datetime.now()}: This is a broadcast."\n).update(id="broadcast_box")\n```\n\nWith the `quickstart/` path open in a browser window, watch as the broadcast pushes messages to the page.\n\nNow change `.update()` to `.append()` and resend the broadcast a few times. Notice you do not have to reload the page to get this modified behavior.\n\nExcited to learn more?  Be sure to walk through the [tutorial](https://turbo-django.readthedocs.io/en/latest/index.html) and read more about what Turbo can do for you.\n\n## Documentation\nRead the [full documentation](https://turbo-django.readthedocs.io/en/latest/index.html) at readthedocs.io.\n\n\n## Contribute\n\nDiscussions about a Django/Hotwire integration are happening on the [Hotwire forum](https://discuss.hotwired.dev/t/django-backend-support-for-hotwire/1570). And on Slack, which you can join by [clicking here!](https://join.slack.com/t/pragmaticmindsgruppe/shared_invite/zt-kl0e0plt-uXGQ1PUt5yRohLNYcVvhhQ)\n\nAs this new magic is discovered, you can expect to see a few repositories with experiments and demos appear in [@hotwire-django](https://github.com/hotwire-django). If you too are experimenting, we encourage you to ask for write access to the GitHub organization and to publish your work in a @hotwire-django repository.\n\n\n## License\n\nTurbo-Django is released under the [MIT License](https://opensource.org/licenses/MIT) to keep compatibility with the Hotwire project.\n\nIf you submit a pull request. Remember to add yourself to `CONTRIBUTORS.md`!\n',
    'author': 'Nikita Marchant',
    'author_email': 'C4ptainCrunch@github-username.x',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from django.apps import AppConfig


class TurboDjangoConfig(AppConfig):
    """The default AppConfig for admin which does autodiscovery."""
    name = "turbo"

    def ready(self):
        super().ready()
        self.module.autodiscover()
